MutexRegister

Struct MutexRegister 

pub struct MutexRegister<T: Copy + Default> { /* private fields */ }
An shared-memory register, backed by a Mutex.

This object uses a mutex to protect against concurrent memory access. It is not lock-free.

Examples

A simple spinlock.

use std::sync::Arc;
use std::{hint, thread};
use todc_mem::register::{MutexRegister, Register};


let register: Arc<MutexRegister<bool>> = Arc::new(MutexRegister::new());

let register_clone = register.clone();
let thread = thread::spawn(move || {
    register_clone.write(true)
});

while !register.read() {
    hint::spin_loop();
}

thread.join().unwrap();

It is also possible to store larger, more complicated objects.

use todc_mem::register::{MutexRegister, Register};

#[derive(Clone, Copy, Debug, Default, PartialEq)]
enum MyType {
    #[default]
    Nothing,
    Booleans([bool; 100]),
    Numbers([u64; 100]),
}

let register: MutexRegister<MyType> = MutexRegister::new();

assert_eq!(register.read(), MyType::Nothing);

let numbers = MyType::Numbers([42; 100]);
register.write(numbers);
assert_eq!(register.read(), numbers);